/*									tab:4
 * PHY_TEST_SENDER.c - periodically send a packet for testing PHY_RADIO
 *  Four groups of packets:
 *  1) Fixed length with TST_PKT_INTERVAL b/w each two packets
 *  2) Variable length with TST_PKT_INTERVAL b/w two packets
 *  3) Fixed length without any delay b/w each two packets (back-to-back)
 *  4) Variable length without any delay b/w two pakcets (back-to-back)
 *  
 *  Transmission time of each packet is measured by txTickCount, and is 
 *  sent in next packet.
 *
 * Authors:   Wei Ye (USC/ISI)
 * History:   created 7/5/2002
 *
 */

#include "tos.h"
#include "include/phy_test_msg.h"
#include "PHY_TEST_SENDER.h"

/* num of pkts in each group, max value is 255 */
#ifndef TST_NUM_PKTS
#define TST_NUM_PKTS 100
#endif

/* packet interval in ms */
#ifndef TST_PKT_INTERVAL
#define TST_PKT_INTERVAL 10
#endif

/* pause between two groups, in ms */
#ifndef TST_GRP_INTERVAL
#define TST_GRP_INTERVAL 1000
#endif

typedef enum {
	FIX_LEN_INTVL,
	VAR_LEN_INTVL,
	FIX_LEN_NO_INTVL,
	VAR_LEN_NO_INTVL,
	STOP
} TxMode;

#define TOS_FRAME_TYPE TEST_frame
TOS_FRAME_BEGIN(TEST_frame) {
	TxMode txMode;
	unsigned short tickCount;  // counter on clock ticks
	unsigned char numTx;  // number of transmitted pkts
	unsigned char txTickCount;  // counter for number of ticks used in tx
	unsigned char numTxTicks;  // number of ticks used in last tx
	unsigned char pktLen;
	unsigned char baseLen;
	unsigned char mask;
	char validLen;
	AppPkt data; 		// packet to be sent
}
TOS_FRAME_END(TEST_frame);


TOS_TASK(sendPkt) {
	// construct a new message
	VAR(data).hdr.seqNo = VAR(numTx);  // record sequence number
	VAR(data).data[0] = VAR(numTxTicks); // time used in tx last pkt
	// select a packet length
	if (VAR(txMode) == FIX_LEN_INTVL || VAR(txMode) == FIX_LEN_NO_INTVL) {
		VAR(pktLen) = (unsigned char)MAX_PKT_LEN;
	} else {
		VAR(pktLen) = VAR(baseLen) + 
			(unsigned char)(TOS_CALL_COMMAND(TEST_SUB_NEXT_RAND)() & VAR(mask));
	}	
	VAR(txTickCount) = 0;  // start counting time to send the pkt
	TOS_CALL_COMMAND(TEST_SUB_TX_PKT)(&VAR(data), VAR(pktLen));
}


char TOS_COMMAND(TEST_INIT)() {
	TOS_CALL_COMMAND(TEST_SUB_INIT)();		// initialize lower components
	TOS_CALL_COMMAND(TEST_SUB_RAND_INIT)();
	VAR(numTx) = 0;
	VAR(numTxTicks) = 0;
	VAR(pktLen) = (unsigned char)MAX_PKT_LEN;
	if (VAR(pktLen) < 14 || VAR(pktLen) > 250) {
		VAR(validLen) = 0;
	} else {
		VAR(validLen) = 1;
		if (VAR(pktLen) >= 14 && VAR(pktLen) < 22) {
			VAR(mask) = 0x7;
		} else if (VAR(pktLen) >= 22 && VAR(pktLen) < 38) {
			VAR(mask) = 0xf;
		} else if (VAR(pktLen) >= 38 && VAR(pktLen) < 70) {
			VAR(mask) = 0x1f;
		} else if (VAR(pktLen) >= 70 && VAR(pktLen) < 134) {
			VAR(mask) = 0x3f;
		} else if (VAR(pktLen) >= 134 && VAR(pktLen) <= 250) {
			VAR(mask) = 0x7f;
		}
		VAR(baseLen) = VAR(pktLen) - VAR(mask);
	}
	return 1;
}


char TOS_COMMAND(TEST_START)() {
	// initialize and start clock
	VAR(txMode) = FIX_LEN_INTVL;
	VAR(tickCount) = TST_PKT_INTERVAL;
	if (VAR(validLen)) {
		TOS_CALL_COMMAND(TEST_SUB_CLOCK_INIT)(tick1000ps); // 1ms resolution
	}
	return 1;
}


void TOS_EVENT(TEST_CLOCK_FIRE)() {
	// clock event handler
	VAR(txTickCount)++;  // measure of transmission time
	if (VAR(tickCount) > 0) {
		VAR(tickCount)--;
		if (VAR(tickCount) == 0) {
			TOS_POST_TASK(sendPkt);
		}
	}
}	


char TOS_EVENT(TEST_SUB_TX_PKT_DONE)(void* msg) {
	TOS_CALL_COMMAND(TEST_LEDr_tog)();
	VAR(numTx)++;
	VAR(numTxTicks) = VAR(txTickCount) + 1;
	if (VAR(numTx) < TST_NUM_PKTS) {
		if (VAR(txMode) == FIX_LEN_NO_INTVL || VAR(txMode) == VAR_LEN_NO_INTVL) {
			// send next pkt without delay
			TOS_POST_TASK(sendPkt);
		} else {
			VAR(tickCount) = TST_PKT_INTERVAL;  // start timer for next packet
		}
	} else {
		VAR(txMode)++;
		if (VAR(txMode) == STOP) { // test is done
			// turn on LEDs to show final results
			TOS_CALL_COMMAND(TEST_LEDr_on)();
			TOS_CALL_COMMAND(TEST_LEDg_on)();
			TOS_CALL_COMMAND(TEST_LEDy_on)();
			return 1;
		} else {
			VAR(numTx) = 0;
			VAR(tickCount) = TST_GRP_INTERVAL;  // start timer for next group
		}
	}
	return 1;
}


char TOS_EVENT(TEST_SUB_START_SYM_DETECTED)(void) {
	// start symbol detected
	return 1;
}

void* TOS_EVENT(TEST_SUB_RX_PKT_DONE)(void* data, char error)
{
	return data;
}

char TOS_EVENT(TEST_SUB_CHANNEL_IDLE)(void) {
	return 1;
}

char TOS_EVENT(TEST_SUB_CHANNEL_BUSY)(void) {
	return 1;
}

