/*									tab:4
 * Copyright (c) 2002 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 * Authors: Wei Ye
 *          (CRC calculation is based on code from Jason Hill)
 * Date created: 7/1/2002
 * 
 * This is the physical layer that sends and receives a packet
 *   - accept any type and length (<= MAX_PKT_LEN in phy_radio_msg.h) of packet
 *   - sending a packet: encoding and byte spooling
 *   - receiving a packet: decoding, byte buffering
 *   - Optional CRC check
 *   - interface to radio control and physical carrier sense
 *
 */

#include "tos.h"
#include "phy_radio_msg.h"
#include "PHY_RADIO.h"
#include "dbg.h"

// if need to stream each received byte to upper layer,
// uncomment following line or define it in Makefile
// #define PHY_STREAM_BYTE

// Physical layer states
#define IDLE 0
#define RECEIVING 1
#define TRANSMITTING 2
#define TRANSMITTING_LAST 3
#define TRANSMITTING_DONE 4

// buffer states
#define FREE 0
#define BUSY 1

#define TOS_FRAME_TYPE PHY_RADIO_frame
TOS_FRAME_BEGIN(PHY_RADIO_frame) {
	char state;
	char crcEnabled;
	unsigned char pktLength; // pkt length including my header and trailer
	PhyPktBuf buffer1;  // 2 buffers for receiving and processing
	PhyPktBuf buffer2;
	char recvBufState;  // receiving buffer state
	char procBufState;  // processing buffer state
	char* procBufPtr;
	char* sendPtr;
	char* recvPtr;
	char* procPtr;
	unsigned char recvCount;
	unsigned char numEncoded;
	char txBuffer[3];
	unsigned char bufHead;
	unsigned char bufEnd;
	short crc;
}
TOS_FRAME_END(PHY_RADIO_frame);


short calcrc(char *ptr, unsigned char count) {
    short crc;
    char i;
    crc = 0;
    while (count > 0)
    {
        crc = crc ^ ((short)*ptr << 8);
		ptr++;
        for (i = 0; i < 8; i++) {
            if (crc & 0x8000)
                crc = crc << 1 ^ 0x1021;  // << is done before ^
            else
                crc = crc << 1;
        }
		count--;
    }
    return (crc);
}


TOS_TASK(calc_crc) {
	short crc;
	crc = calcrc(VAR(sendPtr), VAR(pktLength) - 2);
	*(short*)(VAR(sendPtr) + VAR(pktLength) - 2) = crc;
}


//TOS_TASK(update_crc) {
void update_crc() {
	char i;
	short tmp;
	tmp = (short)(VAR(recvPtr)[VAR(recvCount) - 1]);
	VAR(crc) = VAR(crc) ^ (tmp << 8);
	for (i = 0; i < 8; i++) {
		if (VAR(crc) & 0x8000)
			VAR(crc) = VAR(crc) << 1 ^ 0x1021;  // << is done before ^
		else
			VAR(crc) = VAR(crc) << 1;
	}
}

	
TOS_TASK(packet_received){
	void* tmp;
	char error;
	unsigned char len;
	len = (unsigned char)VAR(procPtr)[0];
	if (VAR(crcEnabled) && VAR(crc) != *(short*)(VAR(procPtr) + len - 2)) {
		error = 1;
	} else {
		error = 0;
	}
	tmp = TOS_SIGNAL_EVENT(PHY_RX_PACKET_DONE)(VAR(procPtr), error);
	if(tmp != 0) {
		if (VAR(recvBufState) == BUSY) { // waiting for a free buffer
			VAR(recvPtr) = (char*)tmp;
			VAR(recvBufState) = FREE;  // can start receive now
		} else {
			VAR(procBufPtr) = (char*)tmp;
			VAR(procBufState) = FREE;
		}
	}
}


TOS_TASK(packet_sent){
    TOS_SIGNAL_EVENT(PHY_TX_PACKET_DONE)(VAR(sendPtr));
}


char TOS_COMMAND(PHY_INIT)(){
    VAR(state) = IDLE;
	VAR(crcEnabled) = 1;  // default is to enable CRC check
	VAR(recvPtr) = (char*)&VAR(buffer1);
	VAR(procBufPtr) = (char*)&VAR(buffer2);
	VAR(recvBufState) = FREE;
	VAR(procBufState) = FREE;
    TOS_CALL_COMMAND(PHY_SUB_RADIO_INIT)();
    return 1;
} 


char TOS_COMMAND(PHY_RADIO_IDLE)(){
	return TOS_CALL_COMMAND(PHY_SUB_RADIO_IDLE)();
}


char TOS_COMMAND(PHY_RADIO_SLEEP)(){
	return TOS_CALL_COMMAND(PHY_SUB_RADIO_SLEEP)();
}


void TOS_COMMAND(PHY_CRC_ENABLE)(){
	VAR(crcEnabled) = 1;
}


void TOS_COMMAND(PHY_CRC_DISABLE)(){
	VAR(crcEnabled) = 0;
}


char TOS_COMMAND(PHY_START_CARR_SEN)(unsigned short numBits){
	return TOS_CALL_COMMAND(PHY_SUB_START_CARR_SEN)(numBits);
}


char TOS_COMMAND(PHY_TX_PACKET)(void* packet, unsigned char length){
    if(VAR(state) != IDLE && VAR(state) != RECEIVING) return 0;
	VAR(state) = TRANSMITTING;
	VAR(sendPtr) = (char*)packet;
	((PhyHeader*)VAR(sendPtr))->length = length;  // fill my header field
	VAR(pktLength) = length;
	if (VAR(crcEnabled)) TOS_POST_TASK(calc_crc);
	// encode first byte of the packet
	VAR(bufHead) = 0;
	VAR(bufEnd) = 0;
	TOS_CALL_COMMAND(PHY_SUB_CODEC_ENCODE)(VAR(sendPtr)[0]);
	VAR(numEncoded) = 1;
  	TOS_CALL_COMMAND(PHY_SUB_START_TX)(); // tell radio to start sending
	return 1;
}


char TOS_EVENT(PHY_SUB_CODEC_ENCODE_DONE)(char data1){
	VAR(txBuffer)[VAR(bufEnd)] = data1;
	VAR(bufEnd)++;
	return 1;
}


char TOS_EVENT(PHY_SUB_CODEC_DECODE_DONE)(char data, char error) {
	// one byte is decoded
	if (VAR(recvCount) == 0) {  // first byte is packet length
		if (error == 1) {
			VAR(state) = IDLE;
			TOS_CALL_COMMAND(PHY_RADIO_IDLE)();
			// signal received an erroneous packet with NULL buffer
			// unknown length (0) and error flag setting to 1
			TOS_SIGNAL_EVENT(PHY_RX_PACKET_DONE)(NULL, 1);
			return 0;
		}
		VAR(pktLength) = (unsigned char)data;
		VAR(crc) = 0;
	}
	VAR(recvPtr)[VAR(recvCount)] = data;
	VAR(recvCount)++;
#ifdef PHY_STREAM_BYTE
	// pass data for upper layer to stream bytes
	TOS_SIGNAL_EVENT(PHY_RX_BYTE_DONE)(data);
#endif
	if (VAR(recvCount) < VAR(pktLength) - 1) {
		//TOS_POST_TASK(update_crc);  // progressive CRC calculation
		update_crc();
	} else if (VAR(recvCount) == VAR(pktLength)) { // Rx packet done
		VAR(procPtr) = VAR(recvPtr);
		if (VAR(procBufState) == FREE) {  // have a free buffer, use it now
			VAR(recvPtr) = VAR(procBufPtr);
			VAR(recvBufState) = FREE;
		} else {  // no buffer to use for Rx
			VAR(recvBufState) = BUSY;
		}
		VAR(procBufState) = BUSY;
		VAR(state) = IDLE;
		TOS_CALL_COMMAND(PHY_SUB_CODEC_INIT)();
		TOS_CALL_COMMAND(PHY_RADIO_IDLE)();
	    TOS_POST_TASK(packet_received);
    }
    return 1;
}


char TOS_EVENT(PHY_SUB_TX_BYTE_READY)() {
	// radio asks a byte to transmit
	if(VAR(state) == TRANSMITTING) {
		TOS_CALL_COMMAND(PHY_SUB_TX_NEXT_BYTE)(VAR(txBuffer)[VAR(bufHead)]);
		VAR(bufHead)++;
		//now check if that was the last byte
		if (VAR(bufHead) == VAR(bufEnd)) {
			VAR(bufHead) = 0;
			VAR(bufEnd) = 0;
			if (VAR(numEncoded) < VAR(pktLength)) {
				TOS_CALL_COMMAND(PHY_SUB_CODEC_ENCODE)(VAR(sendPtr)[VAR(numEncoded)]);
				VAR(numEncoded)++;
			} else {
				TOS_CALL_COMMAND(PHY_SUB_CODEC_ENCODE_FLUSH)();
				if (VAR(bufHead) == VAR(bufEnd)) {
					// tx is done
					VAR(state) = TRANSMITTING_LAST;
				}
			}
		}	
	} else if (VAR(state) == TRANSMITTING_LAST) {
		VAR(state) = TRANSMITTING_DONE;
	} else if (VAR(state) == TRANSMITTING_DONE) {
		VAR(state) = IDLE;
		TOS_CALL_COMMAND(PHY_RADIO_IDLE)();
		TOS_POST_TASK(packet_sent);
	}
    return 1; 
}


char TOS_EVENT(PHY_SUB_START_SYM_DETECTED)() {
	if (VAR(state) == IDLE && VAR(recvBufState) == FREE) {
		VAR(state) = RECEIVING;
		VAR(recvCount) = 0;
		TOS_SIGNAL_EVENT(PHY_START_SYM_DETECTED)();
		return 1;
	}
	return 0;
}


char TOS_EVENT(PHY_SUB_RX_BYTE_DONE)(char data) {
	if (VAR(state) == RECEIVING) {
		TOS_CALL_COMMAND(PHY_SUB_CODEC_DECODE)(data);
		return 1;
	}
	return 0;
}

