/* SMAC.c                                                      tab:4
 *
 * Copyright (c) 2001 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 *
 * Authors:  Wei Ye (USC/ISI)
 *
 * This module implements Sensor-MAC (S-MAC)
 * http://www.isi.edu/scadds/papers/smac_infocom.pdf
 *
 * It has the following functions.
 *  1) Low-duty-cycle operation on radio -- periodic listen and sleep
 *     Option to disable sleep cycles -- define SMAC_NO_SLEEP_CYCLE in Makefile
 *  2) Broadcast only uses CSMA
 *  3) Many features for unicast
 *     - RTS/CTS for hidden terminal problem
 *     - fragmentation support for a long message
 *       A long message is divided (by upper layer) into multiple fragments.
 *       The RTS/CTS reserves the medium for the entire message.
 *       ACK is used for each fragment for immediate error recovery.
 *     - Node goes to sleep when its neighbors are talking to other nodes.
 *
 */


extern short TOS_LOCAL_ADDRESS;

#include "tos.h"
#include "smac_msg.h"
#include "SMAC.h"

//#define SMAC_TEST
//#define SP_DEBUG
//#define SP_DEBUG_STATE
//#define SMAC_DEBUG
//#define FULLPC_SMAC_DEBUG
//#define SMAC_PERFORMANCE

/* User-adjustable MAC parameters
 *--------------------------------
 * The default values can be overriden in Each application's Makefile
 * SMAC_MAX_NUM_NEIGHB: maximum number of neighbors.
 * SMAC_MAX_NUM_SCHED: maximum number of different schedules.
 * SMAC_DUTY_CYCLE: duty cycle in percentage. It controls the length of sleep 
 *   interval.
 * SMAC_RETRY_LIMIT: maximum number of RTS retries for sending a single message.
 * SMAC_EXTEND_LIMIT: maximum number of times to extend Tx time when ACK timeout
     happens.
 */
#ifndef SMAC_MAX_NUM_NEIGHB
#define SMAC_MAX_NUM_NEIGHB 20
#endif

#ifndef SMAC_MAX_NUM_SCHED
#define SMAC_MAX_NUM_SCHED 4
#endif

#ifndef SMAC_DUTY_CYCLE
#define SMAC_DUTY_CYCLE 10
#endif

#ifndef SMAC_RETRY_LIMIT
#define SMAC_RETRY_LIMIT 3
#endif

#ifndef SMAC_EXTEND_LIMIT
#define SMAC_EXTEND_LIMIT 3
#endif

/* Internal MAC parameters
 *--------------------------
 * Do NOT change them unless for tuning S-MAC
 * SLOTTIME: time of each slot in contention window, in ms. It should be large
 *   enough to receive the whole start symbol.
 * DIFS: DCF interframe space (from 802.11), in ms. It is used at the beginning
 *   of each contention window. It's the minmum time to wait to start a new 
 *   transmission.
 * SIFS: short interframe space (from 802.11), in ms. It is used before sending
 *   an CTS or ACK packet. It takes care of the processing delay of each pkt.
 * SYNC_CW: number of slots in the sync contention window, must be 2^n - 1 
 * DATA_CW: number of slots in the data contention window, must be 2^n - 1
 * GUARDTIME: guard time at the end of each listen interval, in ms.
 * SYNC_PERIOD: period to send a sync pkt, in ms.
 * SRCH_CYCLES_LONG: wake up for an entire SYNC_PERIOD in every n SYNC_PERIODS
 * SRCH_CYCLES_SHORT: if I have no neighbor, search more aggressively
 */
#define DIFS 10
#define SIFS 5
#define EIFS 50
#define SLOTTIME 1
//#define SYNC_CW 31
//#define DATA_CW 63
//#define GUARDTIME 4
#define SYNC_CW 15
#define DATA_CW 31
#define GUARDTIME 5
#define SYNC_PERIOD 10000
#define SRCH_CYCLES_LONG 12
#define SRCH_CYCLES_SHORT 3

/* Physical layer parameters
 *---------------------------
 * Based on the parameters from PHY_RADIO and RADIO_CONTROL
 * CLOCK_RES: clock resolution in ms. 
 * BANDWIDTH: bandwidth (bit rate) in kbps. Not directly used.
 * LISTEN_RATE: sampling rate of radio in listen mode (carrier sense)
 * PRE_PKT_BYTES: number of extra bytes transmitted before each pkt. It equals
 *   preamble + start symbol + sync bytes.
 * ENCODE_RATIO: output/input ratio of the number of bytes of the encoding
 *  scheme. In Manchester encoding, 1-byte input generates 2-byte output.
 * PROC_DELAY: processing delay of each packet in physical and MAC layer, in ms
 */
#define CLOCK_RES 1
#define BANDWIDTH 20
#define LISTEN_RATE 20    /* same as BANDWIDTH */
#define PRE_PKT_BYTES 5
#define ENCODE_RATIO 2    /* Manchester encoding has 2x overhead */
#define PROC_DELAY 1

/* MAC states
 *-------------
 * SLEEP: radio is turned off, can't Tx or Rx
 * IDLE: radio in idle mode, will go to Rx if start symbol is deteded.
 *   Can start Tx in this state
 * CARR_SENSE: carrier sense. Do it before initiate a Tx
 * BACKOFF - medium is busy, and cannot Tx
 * WAIT_CTS - just sent RTS, and is waiting for CTS
 * WAIT_DATA - just sent CTS, and is waiting for DATA
 * WAIT_ACK - just sent DATA, and is waiting for ACK
 * TX_NEXT_FRAG - just send one fragment, waiting to tx next fragment
 */
#define SLEEP 0
#define IDLE 1
#define CARR_SENSE 2
#define BACKOFF 3
#define WAIT_CTS 4
#define WAIT_DATA 5
#define WAIT_ACK 6
#define TX_NEXT_FRAG 7
#define DATA_SENSE1 8
#define DATA_SENSE2 9

// radio states
#define RADIO_SLEEP 0
#define RADIO_IDLE 1
#define RADIO_RX 2
#define RADIO_TX 3

// how to send a pkt: broadcast or unicast
#define BCAST_DATA 0
#define SEND_SYNC 1
#define SEND_RTS 2
#define SEND_CTS 3
#define SEND_DATA 4
#define SEND_ACK 5

// MAC packet types
#define DATA_PKT 0
#define RTS_PKT 1
#define CTS_PKT 2
#define ACK_PKT 3
#define SYNC_PKT 4

// how to enter sleep mode
#define TRY 0
#define FORCE 1

// data type definitions
// note: pkt formats are defined in tos/include/smac_msg.h
typedef struct {
	unsigned char numNodes;  // number of nodes on this schedule
	char txSync;   // flag indicating need to send sync
	char txData;   // flag indicating need to send data
	unsigned short counter;  // tick counter
	unsigned short syncTime; // counter for sending SYNC
} SchedTable;

typedef struct {
	char state;
	short nodeId;
	unsigned char schedId;
} NeighbList;

#define TOS_FRAME_TYPE SMAC_frame
TOS_FRAME_BEGIN(MAC_SLEEP_frame) {
	// state variables
	char state;			// MAC state
	char opMode;        // mode of operation
	char srchNeighb;    // flag to keep listening for finding possible neighbors
	char schedListen;   // flag indicating the time of scheduled listen
	unsigned char numNeighb;  // number of known neighbors
	NeighbList neighbList[SMAC_MAX_NUM_NEIGHB]; // neighbor list
	unsigned char numSched;  // number of different schedules
	SchedTable schedTab[SMAC_MAX_NUM_SCHED];   // schedule table
	char radioState;  // radio state
	char schedState;  // schedule state: first, second schedule...
	
	// timing variables
	unsigned long clockTime;    // clock time in mili-second
	unsigned long timeStamp;   // time stamp of each incoming packet
	unsigned long extTimeStamp;   // time stamp for timeSync
	unsigned short syncTime;  // time of listen/contention interval for sync pkt
	unsigned short dataTime;  // time of listen/contention interval for data pkt
	unsigned short listenTime;  // time of listening interval
	unsigned short sleepTime;   // time of sleeping interval 
	unsigned short period;      // period of my schedule =listenTime + sleepTime
	unsigned char numSync;      // used for set/clear find neighbor flag
	unsigned short durDataPkt;   // duration (tx time needed) of data packet 
	unsigned short durCtrlPkt;   // duration (tx time needed) of control packet
	unsigned short durSyncPkt;   // duration (tx time needed) of sync packet
	unsigned short timeWaitCtrl; // time to wait for a control packet
	unsigned short nav;	// network allocation vector. nav>0 -> medium busy
	unsigned short neighbNav; // track neighbors' NAV while I'm sending/receiving
	unsigned short geneTime;    // generic timer
	unsigned char txDelay;	    // timer for carrier sense
	unsigned short adapTime;    // timer for adaptive listen
#ifdef SMAC_NO_SLEEP_CYCLE
	unsigned char retryTime;
#endif
	
	// Variables for Tx
	char txRequest;          // if I have accept a tx request;
	char howToSend;		     // what action to take for tx
	short sendAddr;		     // node that I'm sending data to
	unsigned char dataSched; // current schedule I'm talking to
	unsigned char syncSched; // current schedule I'm talking to
	unsigned char numRetry;	 // number of RTS tries for a data pkt
	unsigned char numExtend; // number of Tx time extensions when ACK timeout
	unsigned char numBcast;  // number of times needed to broadcast a packet
	unsigned char numFrags;  // number of fragments in this transmission
	unsigned char txFragCount; // number of transmitted fragments
	unsigned char txPktLen;  // length of data pkt to be transmitted
	MACHeader* dataPkt; // pointer to tx data pkt, only access MAC header
	MACCtrlPkt ctrlPkt;	  // MAC control packet
	MACSyncPkt syncPkt;   // MAC sync packet
	
	// Variables for Rx
	short recvAddr;		      // node that I'm receiving data from
	unsigned char lastRxFrag;  // fragment no of last received fragments

#ifdef SMAC_DEBUG
	char numLenErr;
	char numCrcErr;
	char numSlpPkt;
	char numCTStimeout;
	char numDATAtimeout;
	char numACKtimeout;
	char numSleeps;
	char numWakeups;
#endif
#ifdef SMAC_PERFORMANCE
	char cntRadioTime;
	RadioTime radioTime;
#endif
}
TOS_FRAME_END(SMAC_frame);

// update my NAV
#define UPDATE_NAV(duration) { \
	if (VAR(nav) < duration) VAR(nav) = duration; \
}

// track my neighbors' NAV
#define TRACK_NAV(duration) { \
	if (VAR(neighbNav) < duration) VAR(neighbNav) = duration; \
}


#ifdef SP_DEBUG
void ud_byte(char data) 
{
  sbi(USR, TXC);
  outp(data, UDR); 
}

void ud_init()
{
  outp(12, UBRR);
  inp(UDR); 
  outp(0x08,UCR);
  sei();
}
#endif


char TOS_COMMAND(MAC_INIT)(char opMode) {
	unsigned char i;
#ifdef SP_DEBUG
	ud_init();
#endif
	// initialize constants
	VAR(durCtrlPkt) = (PRE_PKT_BYTES + sizeof(MACCtrlPkt) * ENCODE_RATIO)
						* 8 / BANDWIDTH + 1;
#ifdef SP_DEBUG
//	ud_byte(VAR(durCtrlPkt));
#endif
	VAR(durSyncPkt) = (PRE_PKT_BYTES + sizeof(MACSyncPkt) * ENCODE_RATIO)
						* 8 / BANDWIDTH + 1;
	VAR(syncTime) = DIFS + SLOTTIME * SYNC_CW + VAR(durSyncPkt) + GUARDTIME;
//	VAR(dataTime) = DIFS + SLOTTIME * DATA_CW + VAR(durCtrlPkt) + GUARDTIME;
	// added time for overhearing CTS so that can do adaptive listen
	VAR(dataTime) = DIFS + SLOTTIME * DATA_CW + VAR(durCtrlPkt) + 
					PROC_DELAY + SIFS + VAR(durCtrlPkt) + GUARDTIME;
	VAR(listenTime) = VAR(syncTime) + VAR(dataTime);
	VAR(period) = VAR(listenTime) * 100 / SMAC_DUTY_CYCLE + 1;
	VAR(sleepTime) = VAR(period) - VAR(listenTime);
	// time to wait for CTS or ACK
	VAR(timeWaitCtrl) = PROC_DELAY + SIFS + VAR(durCtrlPkt) + PROC_DELAY;
#ifdef SP_DEBUG
//	ud_byte(VAR(listenTime));
//	ud_byte(VAR(timeWaitCtrl));
#endif

	// initialize state variables
	VAR(state) = IDLE;
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
//	VAR(opMode) = LOWDUTYCYCLE;
	VAR(radioState) = RADIO_IDLE;

	// initialize neighbor list
	VAR(numNeighb) = 0;  // number of known neighbors
	for (i = 0; i < SMAC_MAX_NUM_NEIGHB; i++) {
		VAR(neighbList)[i].state = 0;  // invalid or dead
	}

    // initialize schedule table
	for (i = 0; i < SMAC_MAX_NUM_SCHED; i++) {
		VAR(schedTab)[i].numNodes = 0;
	}
	
#ifndef SMAC_NO_SLEEP_CYCLE
	// choose a tentative schedule, but don't broadcast until
	// listening for a whole SYNC_PERIOD.
	VAR(schedState) = 1;  // this is my first schedule
	VAR(numSched) = 1;
	VAR(schedTab)[0].numNodes = 1;  // I'm the only one on this schedule
	VAR(schedTab)[0].syncTime = SYNC_PERIOD;
	VAR(schedTab)[0].txData = 0;
	VAR(schedTab)[0].txSync = 0;
	VAR(schedTab)[0].counter = VAR(listenTime);  

	// Don't go to sleep for the first SYNC period.
	VAR(srchNeighb) = 1;
	VAR(numSync) = 1;
	VAR(schedListen) = 1;
#endif

	// initialize timing variables
	VAR(clockTime) = 0;
	VAR(nav) = 0;
	VAR(neighbNav) = 0;
	VAR(txDelay) = 0;
	VAR(adapTime) = 0;
#ifdef SMAC_NO_SLEEP_CYCLE
	VAR(retryTime) = 0;
#endif
	
	// initialize Tx variables
	VAR(txRequest) = 0;
	
	// fill in fixed portion of control packet
	VAR(ctrlPkt).fromAddr = TOS_LOCAL_ADDRESS;
	// fill in fixed portion of sync packet
	VAR(syncPkt).fromAddr = TOS_LOCAL_ADDRESS;
	VAR(syncPkt).type = SYNC_PKT;
	VAR(syncPkt).state = VAR(schedState);

#ifdef SMAC_DEBUG
	VAR(numLenErr) = 0;
	VAR(numCrcErr) = 0;
	VAR(numCTStimeout) = 0;
	VAR(numDATAtimeout) = 0;
	VAR(numACKtimeout) = 0;
	VAR(numSleeps) = 0;
	VAR(numWakeups) = 0;
#endif
#ifdef SMAC_PERFORMANCE
	VAR(cntRadioTime) = 0;
#endif

	TOS_CALL_COMMAND(MAC_SUB_RAND_INIT)();
	// initialize and start clock
	TOS_CALL_COMMAND(MAC_SUB_CLOCK_INIT)(tick1000ps);  // 1 ms per tick
		
	//TOS_CALL_COMMAND(MAC_SUB_RADIO_IDLE)(); // turn on the radio
	//initialize physical layer
	TOS_CALL_COMMAND(MAC_SUB_PHY_INIT)();

#ifdef FULLPC_SMAC_DEBUG
	printf("MAC Module initialized\n");
#endif
    return 1;
}


uint32_t TOS_COMMAND(MAC_GET_TIME)() {
	return VAR(clockTime);
}


uint16_t TOS_COMMAND(MAC_GET_PERIOD)(){
#ifdef SMAC_NO_SLEEP_CYCLE
	return 0;
#else
	return VAR(period);
#endif
}


void* TOS_COMMAND(MAC_COUNT_RADIO_TIME)(char start) {
#ifdef SMAC_PERFORMANCE
	// start/stop counting time of radio in different states
	if (start) {
		VAR(radioTime).sleepTime = 0;
		VAR(radioTime).idleTime = 0;
		VAR(radioTime).rxTime = 0;
		VAR(radioTime).txTime = 0;
		VAR(cntRadioTime) = 1;
	} else {
		VAR(cntRadioTime) = 0;
	}
	return &VAR(radioTime);
#else
	return 0;
#endif
}
		

char TOS_COMMAND(MAC_SEND_START_PKT)(char start) {
	// this function will be removed in next release
	return 1;
}


void setMySched(void* pkt, unsigned short refTime) {
	// switch my schedule (first entry of schedule table)
	// now only happens when the first neighbor is found
	MACSyncPkt* packet = (MACSyncPkt*)pkt;
	// follow the schedule in syncPkt
	VAR(schedTab)[0].counter = refTime;  //packet->sleepTime;
	VAR(schedTab)[0].syncTime = 0;
	VAR(schedTab)[0].txSync = 1;  // need to broadcast my schedule
	VAR(schedTab)[0].numNodes++;  // 2 nodes on this schedule now
	VAR(schedState)++;
	// fill in the field in my syncPkt
	VAR(syncPkt).state = VAR(schedState);
	// add my first neighbor
	VAR(neighbList)[0].state = packet->state;
	VAR(neighbList)[0].nodeId = packet->fromAddr;
	VAR(neighbList)[0].schedId = 0;
	VAR(numNeighb) = 1;
}


void TOS_COMMAND(MAC_POWER)(char mode) {
	// only available in NO_SLEEP_CYCLE mode
#ifdef SMAC_NO_SLEEP_CYCLE
    if (mode == 0) {
		sleep(FORCE);
	} else {
		wakeup();
	}
#endif
}


void sleep(char manner) {
	// try or force to enter sleep mode
	// if manner == FORCE, turn off radio immediately
	// if manner == TRY, check other status. May stay in idle.
	if (VAR(state) == SLEEP) return;
	if (manner == FORCE || (VAR(radioState) == RADIO_IDLE && 
		VAR(srchNeighb) == 0 && VAR(schedListen) == 0)) {
		TOS_CALL_COMMAND(MAC_SUB_RADIO_SLEEP)();  // turn off the radio
		VAR(state) = SLEEP;
		VAR(radioState) = RADIO_SLEEP;
#ifdef SMAC_DEBUG
		VAR(numSleeps)++;
#endif
#ifdef SMAC_TEST
		TOS_SIGNAL_EVENT(MAC_SLEEP)();  // signal upper layer
#endif
	} else {
		VAR(state) = IDLE;
	}
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
}


void wakeup() {
	// wake up from sleep, turn on radio and signal upper layer
	if (VAR(state) != SLEEP) return;
	TOS_CALL_COMMAND(MAC_SUB_RADIO_IDLE)();  // turn on the radio
	VAR(state) = IDLE;
	VAR(radioState) = RADIO_IDLE;
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
#ifdef SMAC_DEBUG
	VAR(numWakeups)++;
#endif
#ifdef SMAC_TEST
	TOS_SIGNAL_EVENT(MAC_WAKEUP)();  // signal upper layer
#endif
}


void tryToSend() {
	// try to send a buffered packet
	unsigned short backoffSlots, listenBits;
	if (VAR(state) == IDLE && VAR(nav) == 0 && VAR(neighbNav) == 0) {
		if (VAR(sendAddr) == TOS_BCAST_ADDR) {
			VAR(howToSend) = BCAST_DATA;
		} else {
			VAR(howToSend) = SEND_RTS;
		}
		VAR(state) = CARR_SENSE;
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
		backoffSlots = TOS_CALL_COMMAND(MAC_SUB_NEXT_RAND)() & DATA_CW;
		listenBits = (DIFS + SLOTTIME * backoffSlots) * LISTEN_RATE;
		TOS_CALL_COMMAND(MAC_SUB_START_CARR_SEN)(listenBits);
	}
	return;
}
		

char TOS_COMMAND(MAC_BCAST_MSG)(void* data, unsigned char length) {
#ifndef SMAC_NO_SLEEP_CYCLE
	unsigned char i;
	if (VAR(numNeighb) == 0) return 0;
#endif
	// Don't accept Tx request if I have already accepted a request
	if (VAR(txRequest) != 0 || data == 0 || length == 0 || length > MAX_PKT_LEN)
		return 0;
	VAR(txRequest) = 1;
	VAR(dataPkt) = (MACHeader*)data;
	VAR(txPktLen) = length;
	VAR(sendAddr) = TOS_BCAST_ADDR;
	// fill in MAC header fields
	VAR(dataPkt)->type = DATA_PKT;  // data pkt
	VAR(dataPkt)->toAddr = TOS_BCAST_ADDR;
	VAR(dataPkt)->fromAddr = TOS_LOCAL_ADDRESS;
	VAR(dataPkt)->duration = 0;
	VAR(dataPkt)->fragNo = 0;
#ifdef SMAC_NO_SLEEP_CYCLE
	// try to send now
	tryToSend();
#else
	// set flag in each schedule
	VAR(numBcast) = 0;
	for (i = 0; i < SMAC_MAX_NUM_SCHED; i++) {
		if (VAR(schedTab)[i].numNodes > 0) {
			VAR(schedTab)[i].txData = 1;
			VAR(numBcast)++;
		}
	}
#endif
	return 1;
}


char TOS_COMMAND(MAC_UCAST_MSG)(void* data, unsigned char length, short toAddr,
		unsigned char numFrags) {
#ifndef SMAC_NO_SLEEP_CYCLE
	unsigned char i, schedId = SMAC_MAX_NUM_SCHED;
#endif
	// Don't accept Tx request if I have already accepted a request
	if (VAR(txRequest) != 0 || data == 0 ||	length == 0 ||
		length > MAX_PKT_LEN || numFrags == 0)
		return 0;
#ifndef SMAC_NO_SLEEP_CYCLE
	for (i = 0; i < SMAC_MAX_NUM_NEIGHB; i++) {
		if (VAR(neighbList)[i].state > 0 &&
			VAR(neighbList)[i].nodeId == toAddr) { // a known neighbor
			schedId = VAR(neighbList)[i].schedId;
			break;
		}
	}
	if (schedId == SMAC_MAX_NUM_SCHED) return 0;  // toAddr is unknow
#endif
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: got new unicast pkt to %x...\n", toAddr);
#endif
	VAR(txRequest) = 1;
   	VAR(dataPkt) = (MACHeader*)data;
	VAR(sendAddr) = toAddr;
	VAR(txPktLen) = length;
	VAR(numFrags) = numFrags;
	VAR(txFragCount) = 0;
	VAR(numRetry) = 0;
	VAR(numExtend) = 0;
	// calculate duration of data packet/fragment
	VAR(durDataPkt) = (PRE_PKT_BYTES + length * ENCODE_RATIO) 
					* 8 / BANDWIDTH + 1;
	// fill in MAC header fields except duration
	VAR(dataPkt)->type = DATA_PKT;  // data pkt
	VAR(dataPkt)->toAddr = VAR(sendAddr);
	VAR(dataPkt)->fromAddr = TOS_LOCAL_ADDRESS;
	VAR(dataPkt)->fragNo = VAR(txFragCount);
#ifdef SMAC_NO_SLEEP_CYCLE
	// try to send now
	tryToSend();
#else
	// set flag for data transmission
	VAR(dataSched) = schedId;
	VAR(schedTab)[schedId].txData = 1;
#endif
	return 1;
}


char TOS_COMMAND(MAC_TX_NEXT_FRAG)(void* data) {
	// Send subsequent fragments
	if (VAR(state) != TX_NEXT_FRAG || data == 0) return 0;
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: got new fragment from upper layer.\n");
#endif
   	VAR(dataPkt) = (MACHeader*)data;
	// fill in MAC header fields except duration
	VAR(dataPkt)->type = DATA_PKT;  // data pkt
	VAR(dataPkt)->toAddr = VAR(sendAddr);
	VAR(dataPkt)->fromAddr = TOS_LOCAL_ADDRESS;
	VAR(dataPkt)->fragNo = VAR(txFragCount);
	if (VAR(neighbNav) >= (SIFS + VAR(durDataPkt) + VAR(timeWaitCtrl))) {
		// schedule to send this fragment
		VAR(howToSend) = SEND_DATA;
		VAR(txDelay) = SIFS;
	} // else will retry when neighbNav timeout
	return 1;
}

char TOS_COMMAND(MAC_CLR_TX_BUF)() {
	// force MAC to drop currently buffered tx packet
	// so that it will accept a new one
#ifndef SMAC_NO_SLEEP_CYCLE
	unsigned char i;
	for (i = 0; i < SMAC_MAX_NUM_SCHED; i++) {
		VAR(schedTab)[i].txData = 0;
	}
#endif
	VAR(txRequest) = 0;	
	return 1;
}

void startBcast() {
	// broadcast data directly; don't use RTS/CTS
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: broadcasting packet...\n");
#endif
	TOS_CALL_COMMAND(MAC_SUB_TX_PACKET)(VAR(dataPkt), VAR(txPktLen));
	VAR(radioState) = RADIO_TX;
}


TOS_TASK(ucastMsgDone) {
	TOS_SIGNAL_EVENT(MAC_UCAST_MSG_DONE)(VAR(txFragCount));
}


void txMsgDone() {
	// prepare to tx next msg
	VAR(txRequest) = 0;
#ifdef SMAC_NO_SLEEP_CYCLE
	VAR(state) = IDLE;
#else
	VAR(schedTab)[VAR(dataSched)].txData = 0;
	sleep(TRY);
#endif
	TOS_POST_TASK(ucastMsgDone);
}


char TOS_EVENT(MAC_SUB_TX_PACKET_DONE)(void* packet) {
	char pktType;
	VAR(radioState) = RADIO_IDLE;
	if (packet == 0) return 0;
	pktType = *((char*)packet + sizeof(PhyHeader));
	switch (pktType) {  // the type field
	case SYNC_PKT:
		VAR(schedTab)[VAR(syncSched)].txSync = 0;
		VAR(schedTab)[VAR(syncSched)].syncTime = SYNC_PERIOD;
		VAR(numSync)--;
		if (VAR(numSync) == 1) {
			VAR(srchNeighb) = 1;
		} else if (VAR(numSync) == 0) {
			VAR(srchNeighb) = 0;
			if (VAR(numNeighb) == 0) {
				VAR(numSync) = SRCH_CYCLES_SHORT;
			} else {
				VAR(numSync) = SRCH_CYCLES_LONG;
			}
		}
		break;
	case RTS_PKT:
		// track neighbors' NAV, need to clear it if CTS timeout
		TRACK_NAV(((MACCtrlPkt*)packet)->duration);
		// just sent RTS, set timer for CTS timeout
		VAR(geneTime) = VAR(timeWaitCtrl);
		break;
	case CTS_PKT:  // just sent CTS
		// track my neighbors' NAV
		// they update NAV and go to sleep after recv CTS
		TRACK_NAV(((MACCtrlPkt*)packet)->duration);
		// no data timeout, just use neighbors' NAV
		// since they went to sleep, just wait data for the entire time
		break;
	case DATA_PKT:
		if (VAR(howToSend) == BCAST_DATA) {
			VAR(schedTab)[VAR(dataSched)].txData = 0;
#ifdef SMAC_NO_SLEEP_CYCLE
			VAR(state) = IDLE;
			VAR(txRequest) = 0;
			TOS_SIGNAL_EVENT(MAC_BCAST_MSG_DONE)(VAR(dataPkt));
#else
			// broadcast data is done for one schedule
			sleep(TRY);
			VAR(numBcast)--;
			if (VAR(numBcast) == 0) {
				// broadcast is done for all schedules
				VAR(txRequest) = 0;
				TOS_SIGNAL_EVENT(MAC_BCAST_MSG_DONE)(VAR(dataPkt));
			}
#endif
		} else {  // unicast is done
			// waiting for ACK, set timer for ACK timeout
			VAR(geneTime) = VAR(timeWaitCtrl);
		}
		break;
	case ACK_PKT:
		TRACK_NAV(((MACCtrlPkt*)packet)->duration); // in case tx extended.
		break;
	}
	return 1;
}


void tryToResend(unsigned char delay) {
	// try to re-send a packet when CTS or ACK timeout
	if (VAR(numRetry) < SMAC_RETRY_LIMIT) {
		VAR(numRetry)++;
#ifdef SMAC_NO_SLEEP_CYCLE
		VAR(state) = IDLE;
		if (delay == 0) tryToSend();
		else VAR(retryTime) = delay;
#else
		// wait unitl receiver's next wake-up time
		sleep(TRY);
#endif
	} else {
		// reached retry limit, give up Tx
#ifdef FULLPC_SMAC_DEBUG
		printf("MAC: Reached retry limit, give up Tx!\n");
#endif
		txMsgDone(); // with txFragCount = 0;
	}
}


void adaptiveListen() {
	// adaptively wake-up at the end of current transmission. Will try to 
	// send only if the buffered packet is unicast. Since my next-hop 
	// neighbor may not be aware of the Tx of my previous-hop neighbor, 
	// broadcast now is unreliable.
	if (VAR(srchNeighb) == 1 || 
		VAR(schedTab)[0].counter > VAR(dataTime) + VAR(listenTime) ||
		VAR(schedTab)[0].counter < VAR(dataTime)) {
		VAR(adapTime) = VAR(dataTime); // set timer to bring me back to sleep
		if (VAR(state) == SLEEP) wakeup();
		else VAR(state) = IDLE;
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
		if (VAR(txRequest) == 1 && VAR(sendAddr) != TOS_BCAST_ADDR &&
			(VAR(schedTab)[VAR(dataSched)].counter > 
			VAR(dataTime) + VAR(listenTime) ||
			VAR(schedTab)[VAR(dataSched)].counter < VAR(dataTime))) {
			tryToSend();
		}
	} else {
		sleep(TRY);
	}
}


void TOS_EVENT(MAC_SUB_CLOCK_FIRE)() {
	// handle clock event
	unsigned char i, numRemain;
	unsigned short listenBits;
#ifdef SMAC_PERFORMANCE
	if (VAR(cntRadioTime)) {
		if (VAR(radioState) == RADIO_SLEEP)
			VAR(radioTime).sleepTime++;
		else if (VAR(radioState) == RADIO_IDLE)
			VAR(radioTime).idleTime++;
		else if (VAR(radioState) == RADIO_RX)
			VAR(radioTime).rxTime++;
		else if (VAR(radioState) == RADIO_TX)
			VAR(radioTime).txTime++;
	}
#endif
	// advance clock
	VAR(clockTime)++;
	
	// NAV timer
	if (VAR(nav) > 0) {
		VAR(nav)--;
		if (VAR(nav) == 0) {  // my neighbor is done Tx/Rx
//#ifdef SMAC_NO_SLEEP_CYCLE
#if defined SMAC_NO_SLEEP_CYCLE
			wakeup();
			if (VAR(txRequest) == 1) tryToSend();
#elif defined SMAC_NO_ADAPTIVE_LISTEN
			if (VAR(srchNeighb) == 1) wakeup();
#else
			adaptiveListen();
#endif
		}
	}
	// timer to track my neighbor's NAV
	if (VAR(neighbNav) > 0) {
		VAR(neighbNav)--;
		if (VAR(neighbNav) == 0) {  // used up reserved Tx/Rx time
			if (VAR(state) == WAIT_ACK || VAR(state) == TX_NEXT_FRAG) {
				// last tx is not done within reserved time
				tryToResend(0);  // try to resend now
			} else { // could be in sleep, idle or wait_data state
//#ifdef SMAC_NO_SLEEP_CYCLE
#if defined SMAC_NO_SLEEP_CYCLE
				VAR(state) = IDLE;
				if (VAR(txRequest) == 1) tryToSend();
#elif defined SMAC_NO_ADAPTIVE_LISTEN
				sleep(TRY);
#else			
				adaptiveListen();
#endif
			}
		}
	}
	// generic timer
	if (VAR(geneTime) > 0) {
		VAR(geneTime)--;
		if (VAR(geneTime) == 0) {
			if (VAR(state) == DATA_SENSE1) {
				// didn't see a CTS reply to a previous RTS
				VAR(state) = DATA_SENSE2;
				VAR(geneTime) = VAR(timeWaitCtrl);
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
			} else if (VAR(state) == DATA_SENSE2) {
				// didn't see data tx, sender of RTS didn't get a CTS
				VAR(nav) = 0;  // cancel my NAV
#ifdef SMAC_NO_SLEEP_CYCLE
				VAR(state) = IDLE;
				if (VAR(txRequest) == 1) tryToSend();
#else
				sleep(TRY);
#endif
			} else if (VAR(state) == WAIT_CTS) {	// CTS timeout
#ifdef SMAC_DEBUG
				VAR(numCTStimeout)++;
#endif
				VAR(neighbNav) = 0; // neighbors won't follow my reserved time
				tryToResend(EIFS); // wait neighbors in data_sense2
			} else if (VAR(state) == WAIT_ACK) {	// ACK timeout
#ifdef SMAC_DEBUG
				VAR(numACKtimeout)++;
#endif
				if (VAR(numExtend) < SMAC_EXTEND_LIMIT && 
					VAR(neighbNav) > VAR(durDataPkt)) {
					// can extend tx time
#ifdef FULLPC_SMAC_DEBUG
					printf("MAC: ACK timeout. Extend Tx time.\n");
#endif
					VAR(numExtend)++;
					// need to increase neighbNav for one frag
					numRemain = VAR(numFrags) - VAR(txFragCount);
					VAR(neighbNav) = numRemain * VAR(durCtrlPkt) + 
						numRemain * VAR(durDataPkt) +
						(numRemain + numRemain - 1) * (PROC_DELAY + SIFS);
					sendDATA();
				} else if (VAR(neighbNav) >= VAR(durDataPkt) + 
						VAR(timeWaitCtrl)) {
					sendDATA(); // can't extend but still have time to send
					// if can't extend or send, retry when neighbNav timeout
				}
			}
		}
	}
	// update status of each schedule and arrange tx if needed
	for (i = 0; i < SMAC_MAX_NUM_SCHED; i++) {
	  if (VAR(schedTab)[i].numNodes > 0) {
		VAR(schedTab)[i].counter--;
		if (VAR(schedTab)[i].counter == VAR(listenTime)) {
			if (i == 0) VAR(schedListen) = 1; // it's my scheduled wake-up time
			if (VAR(nav) == 0 && VAR(neighbNav) == 0 &&
				(VAR(state) == SLEEP || VAR(state) == IDLE)) {
				if (VAR(state) == SLEEP &&
					(i == 0 || VAR(schedTab)[i].txSync == 1)) {
					wakeup();  // wake up to listen or send sync
				}
				if (VAR(schedTab)[i].txSync == 1) {
					// start carrier sense for sending sync
					VAR(howToSend) = SEND_SYNC;
					VAR(syncSched) = i;
					VAR(state) = CARR_SENSE;
					numRemain = TOS_CALL_COMMAND(MAC_SUB_NEXT_RAND)() & SYNC_CW;
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
					listenBits = (DIFS + SLOTTIME * numRemain) * LISTEN_RATE;
					//(TOS_CALL_COMMAND(MAC_SUB_NEXT_RAND)() & SYNC_CW)) * LISTEN_RATE;
					TOS_CALL_COMMAND(MAC_SUB_START_CARR_SEN)(listenBits);
				}
			}
		} else if (VAR(schedTab)[i].counter == VAR(dataTime)) {
			if (VAR(schedTab)[i].txData == 1 &&
				VAR(nav) == 0 && VAR(neighbNav) == 0 &&
				(VAR(state) == SLEEP || VAR(state) == IDLE)) {
				// schedule sending of data
				if (VAR(state) == SLEEP) wakeup();  // wake up to tx
				VAR(dataSched) = i;
				tryToSend();
			}
		} else if (VAR(schedTab)[i].counter == 0) {
			VAR(schedTab)[i].counter = VAR(period);
			if (i == 0) {  // my scheduled lisen time is over
				VAR(schedListen) = 0;
				if (VAR(state) == IDLE) sleep(TRY); // sleep if I'm idle
			}
		}
		// how long since my last sync
		if (VAR(schedTab)[i].syncTime > 0) {
			VAR(schedTab)[i].syncTime--;
			if (VAR(schedTab)[i].syncTime == 0) {
				VAR(schedTab)[i].txSync = 1;
			}
		}
	  }
	}

	// tx delay timer
	if (VAR(txDelay) > 0) {
		VAR(txDelay)--;
		if (VAR(txDelay) == 0) {  // start sending
			if (VAR(howToSend) == SEND_CTS) {
				sendCTS();
			} else if (VAR(howToSend) == SEND_DATA) {
				sendDATA();
			} else if (VAR(howToSend) == SEND_ACK) {
				sendACK();
			}
		}
	}
	
	// adaptive wake-up timer
	if (VAR(adapTime) > 0) {
		VAR(adapTime)--;
		if (VAR(adapTime) == 0) {
			if (VAR(state) == IDLE) sleep(TRY);
		}
	}

#ifdef SMAC_NO_SLEEP_CYCLE
	if (VAR(retryTime) > 0) {
		VAR(retryTime)--;
		if (VAR(retryTime) == 0) {
			tryToSend();
		}
	}
#endif

#ifdef SMAC_TEST
	// signal upper layer the clock event
	TOS_SIGNAL_EVENT(MAC_CLOCK_FIRE)();
#endif
	return;
}


char TOS_EVENT(MAC_SUB_CHANNEL_BUSY)(void) {
	// physical carrier sense indicate channel busy
	// Do nothing and stay in idle to receive a packet
	// Will sleep at my next sleep time if can't get a packet
	if (VAR(state) == CARR_SENSE) {
		VAR(state) = IDLE;
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
#ifdef SMAC_NO_SLEEP_CYCLE
		VAR(retryTime) = EIFS;
#endif
	}
	return 1;
}


char TOS_EVENT(MAC_SUB_CHANNEL_IDLE)(void) {
	// physical carrier sense indicate channel idle
	// start sending
	if (VAR(state) != CARR_SENSE) return 0;
	if (VAR(howToSend) == SEND_SYNC) {
		sendSYNC();
	} else if (VAR(howToSend) == BCAST_DATA) {
		startBcast();
	} else if (VAR(howToSend) == SEND_RTS) {
		sendRTS();
	}	
	return 1;
}


char TOS_EVENT(MAC_SUB_START_SYM_DETECTED)() {
	VAR(radioState) = RADIO_RX;
	if (VAR(state) == IDLE || VAR(state) == CARR_SENSE) {	
		VAR(state) = BACKOFF;
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
	} else if (VAR(state) == DATA_SENSE2) {
		// data tx started from the sender of RTS
		sleep(FORCE);
	}
	VAR(timeStamp) = VAR(clockTime);
#ifdef SMAC_EXT_TIMESTAMP
	TOS_CALL_COMMAND(MAC_SUB_GET_EXT_TIME32)(&VAR(extTimeStamp));
#endif
#ifdef SMAC_NO_SLEEP_CYCLE
	VAR(retryTime) = 0;
#endif
	return 1;
}


void* TOS_EVENT(MAC_SUB_RX_PACKET_DONE)(void* packet, char error)
{
	char pktType;

	VAR(radioState) = RADIO_IDLE;
	if (VAR(state) == SLEEP) {
		return packet; // if in sleep, reject any pkt
#ifdef SMAC_DEBUG
		VAR(numSlpPkt)++;
#endif
	}

	if (error) {  // if received an erroneous pkt, a sign of collision
#ifdef SMAC_DEBUG
		if (packet == NULL) VAR(numLenErr)++;
		else VAR(numCrcErr)++;
#endif
		if (VAR(state) == BACKOFF) {
#ifdef SMAC_NO_SLEEP_CYCLE
			VAR(state) = IDLE;
			if (VAR(txRequest) == 1) VAR(retryTime) = EIFS;
#else
			sleep(TRY);  // wait until next tx time
#endif
		}
		return packet;
	}
	
	pktType = *((char*)packet + sizeof(PhyHeader));
	// dispatch to actual packet handlers
	if (pktType == DATA_PKT) {
		return handleDATA(packet);
	} else if (pktType == RTS_PKT) {
		handleRTS(packet);
	} else if (pktType == CTS_PKT) {
		handleCTS(packet);
	} else if (pktType == ACK_PKT) {
		handleACK(packet);
#ifndef SMAC_NO_SLEEP_CYCLE
	} else if (pktType == SYNC_PKT) {
		handleSYNC(packet);
#endif
	}
	return packet;
}


void handleRTS(void* pkt) {
	// internal handler for RTS
	MACCtrlPkt* packet;
	if (VAR(state) != IDLE && VAR(state) != BACKOFF) return;
	packet = (MACCtrlPkt*)pkt;
	if (packet->toAddr == TOS_LOCAL_ADDRESS) {
		VAR(recvAddr) = packet->fromAddr;  // remember sender's address
		VAR(lastRxFrag) = 250;
#ifdef FULLPC_SMAC_DEBUG
		printf("MAC: Received RTS from: %x\n", VAR(recvAddr));
#endif
		TRACK_NAV(packet->duration); // track neighbors' NAV
		// schedule sending CTS
		VAR(howToSend) = SEND_CTS;
		VAR(txDelay) = SIFS;
	} else { // packet destined to another node
		// assume sender will get a CTS and tx data
		UPDATE_NAV(packet->duration);
		// keep listening until confirm sender gets a CTS or starts tx data
		VAR(state) = DATA_SENSE1; // wait for a CTS
		VAR(geneTime) = VAR(timeWaitCtrl);
#ifdef SP_DEBUG_STATE
		ud_byte(VAR(state));
#endif
	}
}
		

void handleCTS(void* pkt) {
	// internal handler for CTS
	MACCtrlPkt* packet;
	packet = (MACCtrlPkt*)pkt;
	if (packet->toAddr == TOS_LOCAL_ADDRESS) {
		if (VAR(state) == WAIT_CTS && packet->fromAddr == VAR(sendAddr)) {
			// cancel CTS timer
			VAR(geneTime) = 0;
#ifdef FULLPC_SMAC_DEBUG
			printf("MAC: Received CTS from: %x\n", VAR(sendAddr));
#endif
//			VAR(schedTab)[VAR(dataSched)].txData = 0;
			// schedule sending DATA
			VAR(howToSend) = SEND_DATA;
			VAR(txDelay) = SIFS;
		}
	} else { // packet destined to another node
		if (VAR(state) == DATA_SENSE1 || VAR(state) == DATA_SENSE2) {
			VAR(geneTime) = 0;
		}
		if (VAR(state) == IDLE || VAR(state) == BACKOFF || 
			VAR(state) == DATA_SENSE1 || VAR(state) == DATA_SENSE2) {
			UPDATE_NAV(packet->duration);
			sleep(FORCE);  // avoid overhearing other's packet
		}
	}
}


void* handleDATA(void* pkt) {
	// internal handler for DATA packet
	void* tmp;
	MACHeader* packet = (MACHeader*)pkt;

	// put in time stamp
#ifdef SMAC_EXT_TIMESTAMP
	((PhyPktBuf*)pkt)->info.timestamp = VAR(extTimeStamp);
#else
	((PhyPktBuf*)pkt)->info.timestamp = VAR(timeStamp);
#endif

	if (packet->toAddr == TOS_BCAST_ADDR) {  // broadcast packet
#ifdef FULLPC_SMAC_DEBUG
		printf("MAC: Received broadcast data pkt from: %x\n", packet->fromAddr);
#endif

#ifdef SMAC_NO_SLEEP_CYCLE
		VAR(state) = IDLE;
		if (VAR(txRequest) == 1) tryToSend();
#else
		sleep(TRY);
#endif
		tmp = TOS_SIGNAL_EVENT(MAC_RX_MSG_DONE)(packet);
		return tmp;
	} else if (packet->toAddr == TOS_LOCAL_ADDRESS) {  // unicast packet
		// could receive data in idle, backoff and wait_data state
		if (VAR(state) == WAIT_DATA && packet->fromAddr == VAR(recvAddr)) {
#ifdef FULLPC_SMAC_DEBUG
			printf("MAC: Received unicast data pkt from: %x\n", packet->fromAddr);
#endif
			// Should track neighbors' NAV, in case tx extended
			TRACK_NAV(packet->duration);
			// schedule sending ACK
			VAR(howToSend) = SEND_ACK;
			VAR(txDelay) = SIFS;
			// if fragNo is the same as last one, don't signal duplicate
			if (VAR(lastRxFrag) != packet->fragNo) {
				VAR(lastRxFrag) = packet->fragNo;
				tmp = TOS_SIGNAL_EVENT(MAC_RX_MSG_DONE)(packet);
				return tmp;
			}
		}
	} else { // unicast packet destined to another node
		if (VAR(state) == IDLE || VAR(state) == BACKOFF) {
			UPDATE_NAV(packet->duration);
			sleep(FORCE); // avoid overhearing other's packet
		}
	}
	
	return pkt;
}


void handleACK(void* pkt) {
	// internal handler for ACK packet
	MACCtrlPkt* packet;
	packet = (MACCtrlPkt*)pkt;
	if (packet->toAddr == TOS_LOCAL_ADDRESS) {
		if (VAR(state) == WAIT_ACK && packet->fromAddr == VAR(sendAddr)) {
			// cancel ACK timer
			VAR(geneTime) = 0;
#ifdef FULLPC_SMAC_DEBUG
			printf("MAC: Received ACK from: %x\n", VAR(sendAddr));
#endif
			VAR(txFragCount)++;  // number of successfully transmitted frags

			if (VAR(txFragCount) < VAR(numFrags)) {
					VAR(state) = TX_NEXT_FRAG;
					TOS_SIGNAL_EVENT(MAC_UCAST_FRAG_DONE)(VAR(dataPkt));
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
			} else { // no more fragment, tx done
				VAR(state) = IDLE;
				TOS_SIGNAL_EVENT(MAC_UCAST_FRAG_DONE)(VAR(dataPkt));
				txMsgDone();
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
			}
		}
	} else { // packet destined to another node
		if (VAR(state) == IDLE || VAR(state) == BACKOFF) {
			UPDATE_NAV(packet->duration);
#ifdef SMAC_NO_SLEEP_CYCLE
			VAR(state) = IDLE;
			if (VAR(txRequest) == 1) tryToSend();
#else
			sleep(TRY);
#endif
		}
	}
}


void handleSYNC(void* pkt)
{
	// internal handler for SYNC packet
	MACSyncPkt* packet = (MACSyncPkt*)pkt;
	unsigned char i, nodeId, schedId, oldSchedId;
	unsigned short refTime, rxDelay;
	char timeDiff;
	
	// if carrier sense failed and received a sync pkt, go back to idle
	if (VAR(state) == BACKOFF) {
		VAR(state) = IDLE;
#ifdef SP_DEBUG_STATE
		ud_byte(VAR(state));
#endif
	}
	
	// calculate Rx delay of sync packet
	rxDelay = (unsigned short)(VAR(clockTime) - VAR(timeStamp)) + 
				PRE_PKT_BYTES * 8 / BANDWIDTH;
#ifdef SP_DEBUG
//	ud_byte(rxDelay);
#endif
	// sanity check
	if (rxDelay < VAR(durSyncPkt) - 2 || rxDelay > VAR(durSyncPkt) + 10)
		return;

	refTime = packet->sleepTime - rxDelay;
#ifdef SP_DEBUG
//	timeDiff = VAR(schedTab)[0].counter - refTime;
//	ud_byte(timeDiff);
#endif
	if (VAR(numNeighb) == 0) {
		// follow this schedule if having no other neighbor
		setMySched(packet, refTime);
		return;
	}
	// check if sender is on my neighbor list
	nodeId = SMAC_MAX_NUM_NEIGHB;
	schedId = SMAC_MAX_NUM_SCHED;
	for (i = 0; i < SMAC_MAX_NUM_NEIGHB; i++) {
		if (VAR(neighbList)[i].state > 0 &&
			VAR(neighbList)[i].nodeId == packet->fromAddr) {
			nodeId = i;
			schedId = VAR(neighbList)[i].schedId; // a known neighbor
			break;
		}
	}
	if (nodeId < SMAC_MAX_NUM_NEIGHB) { 
		if (VAR(neighbList)[nodeId].state == packet->state) {
			// update the existing schedule
			VAR(schedTab)[schedId].counter = refTime; //packet->sleepTime;
			return;
		}
	}

	// now it's either a new node or an old node switching to a new schedule
	// check if its schedule is a known one to me
	oldSchedId = schedId;
	schedId = SMAC_MAX_NUM_SCHED;
	for (i = 0; i < SMAC_MAX_NUM_SCHED; i++) {
		if (VAR(schedTab)[i].numNodes > 0) {
			timeDiff = VAR(schedTab)[i].counter - refTime;
			if (timeDiff > -GUARDTIME && timeDiff < GUARDTIME) {
				VAR(schedTab)[i].counter = refTime; // packet->sleepTime;
				VAR(schedTab)[i].numNodes++; // it will follow this schedule
				schedId = i;
				break;
			}
		}
	}
	if (schedId == SMAC_MAX_NUM_SCHED) {  // unknow schedule
		// add an entry to the schedule table
		for (i = 0; i < SMAC_MAX_NUM_SCHED; i++) {
			if (VAR(schedTab)[i].numNodes == 0) { // found an empty entry
				VAR(schedTab)[i].counter = refTime; //packet->sleepTime;
				VAR(schedTab)[i].txSync = 1;  // need send sync
				VAR(schedTab)[i].txData = 0;
				VAR(schedTab)[i].syncTime = 0;
				VAR(schedTab)[i].numNodes = 1; // 1st node following this sched
				schedId = i;
				break;
			}
		}
		VAR(numSched)++;  // increment number of schedules
	}
	
	if (nodeId == SMAC_MAX_NUM_NEIGHB) {  // a new node
		// add it to my neighbor list
		for (i = 0; i < SMAC_MAX_NUM_NEIGHB; i++) {
			if (VAR(neighbList)[i].state == 0) { // found an empty entry
				VAR(neighbList)[i].state = packet->state;
				VAR(neighbList)[i].nodeId = packet->fromAddr;
				VAR(neighbList)[i].schedId = schedId;
				break;
			}
		}
		VAR(numNeighb)++;  // increment number of neighbors
	} else {  // old node switches to a new schedule
		VAR(neighbList)[nodeId].state = packet->state;
		VAR(neighbList)[nodeId].schedId = schedId;
		// decrement number of nodes on old schedule
		VAR(schedTab)[oldSchedId].numNodes--;
		if (VAR(schedTab)[oldSchedId].numNodes == 0)
			VAR(numSched)--;
	}
#ifdef SP_DEBUG
//	ud_byte(VAR(numNeighb));
#endif
}


void sendRTS() {
	// construct and send RTS packet
	unsigned char numRemain;
	VAR(ctrlPkt).type = RTS_PKT;
	VAR(ctrlPkt).toAddr = VAR(sendAddr);
	// reserve time for CTS + all fragments + all ACKs
	numRemain = VAR(numFrags) - VAR(txFragCount);
	VAR(ctrlPkt).duration = (numRemain + 1) * VAR(durCtrlPkt) + 
					numRemain * VAR(durDataPkt) + 
					(numRemain + numRemain + 1) * (PROC_DELAY + SIFS);
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: Sending RTS to: %x\n", VAR(sendAddr));
#endif
	// send RTS
	TOS_CALL_COMMAND(MAC_SUB_TX_PACKET)(&VAR(ctrlPkt), sizeof(VAR(ctrlPkt)));
	VAR(state) = WAIT_CTS;
	VAR(radioState) = RADIO_TX;
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
}


void sendCTS() {
	// construct and send CTS
	// input duration is the duration field from received RTS pkt
	VAR(ctrlPkt).type = CTS_PKT;
	VAR(ctrlPkt).toAddr = VAR(recvAddr);
	// should track neighbors' NAV as soon as RTS is received
	VAR(ctrlPkt).duration = VAR(neighbNav) - VAR(durCtrlPkt);
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: Sending CTS to: %x\n", VAR(recvAddr));
#endif
	// send CTS
	TOS_CALL_COMMAND(MAC_SUB_TX_PACKET)(&VAR(ctrlPkt), sizeof(VAR(ctrlPkt)));
	VAR(state) = WAIT_DATA; // prepare to receive DATA
	VAR(radioState) = RADIO_TX;
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
}


void sendDATA() {
	// send a unicast data packet
	// assume all MAC header fields have been filled except the duration
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: Sending DATA to: %x\n", VAR(sendAddr));
#endif
#ifdef SMAC_DEBUG
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 10) = VAR(numLenErr);
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 11) = VAR(numCrcErr);
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 12) = VAR(numSlpPkt);
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 13) = VAR(numCTStimeout);
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 14) = VAR(numDATAtimeout);
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 15) = VAR(numACKtimeout);
	
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 17) = VAR(numSleeps);
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 18) = VAR(numWakeups);
	*((char*)VAR(dataPkt) + sizeof(MACHeader) + 19) = VAR(neighbNav);
#endif
	// neighbNav tracks the time I have left for tx
	VAR(dataPkt)->duration = VAR(neighbNav) - VAR(durDataPkt);
	
	TOS_CALL_COMMAND(MAC_SUB_TX_PACKET)(VAR(dataPkt), VAR(txPktLen));
	VAR(state) = WAIT_ACK;
	VAR(radioState) = RADIO_TX;
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
}


void sendACK() {
	// construct and send ACK
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: Sending ACK to: %x\n", VAR(recvAddr));
#endif
	VAR(ctrlPkt).type = ACK_PKT;
	VAR(ctrlPkt).toAddr = VAR(recvAddr);
	// stick to neighbNav -- should update it when rx data packet
	VAR(ctrlPkt).duration = VAR(neighbNav) - VAR(durCtrlPkt);
	TOS_CALL_COMMAND(MAC_SUB_TX_PACKET)(&VAR(ctrlPkt), sizeof(VAR(ctrlPkt)));
	// MAC stay in WAIT_DATA state until neighbNav becomes zero
	VAR(radioState) = RADIO_TX;
}


void sendSYNC() {
	// construct and send SYNC packet
#ifdef FULLPC_SMAC_DEBUG
	printf("MAC: Broadcasting SYNC packet: %x\n");
#endif
	VAR(syncPkt).sleepTime = VAR(schedTab)[0].counter;
	VAR(syncPkt).seqNo = VAR(numSched);
//	VAR(syncPkt).seqNo = VAR(numNeighb);
	TOS_CALL_COMMAND(MAC_SUB_TX_PACKET)(&VAR(syncPkt), sizeof(VAR(syncPkt)));
	VAR(state) = IDLE;
	VAR(radioState) = RADIO_TX;
#ifdef SP_DEBUG_STATE
	ud_byte(VAR(state));
#endif
}


